#include "impr.h"

/* operations for binary images */

/**
 * impr_negate
 * @dest: destination image
 * @src: source image
 * @size: image size
 *
 * Invert image values.
 * @dest == @src is ok.
 */
void impr_negate(b8 *dest, b8 *src, int size)
{
	int i;
	for (i=0; i<size; i++)
		dest[i] = !src[i];
}


/**
 * impr_dilaten
 * @dest: destination image
 * @src: source image
 * @w: image width
 * @h: image height
 * @conn: connectedness (4 or 8)
 * @times: dilate how many times
 *
 * 4 or 8 connected dilation applied @times times.
 * @dest == @src is ok.
 */
void impr_dilaten(b8 *dest, b8 *src, int w, int h, int conn, int times)
{
	assert(conn == 4 || conn == 8);
	if (conn == 4)
		impr_manhattan_distance((u8*)dest, src, w, h);
	else
		impr_chessboard_distance((u8*)dest, src, w, h);

	impr_binarize(dest, (u8*)dest, w*h, <= times);
}

/**
 * impr_eroden
 * @dest: destination image
 * @src: source image
 * @w: image width
 * @h: image height
 * @conn: connectedness (4 or 8)
 * @times: erode how many times
 *
 * 4 or 8 connected erosion applied @times times.
 * @dest == @src is ok.
 */
void impr_eroden(b8 *dest, b8 *src, int w, int h, int conn, int times)
{
	impr_negate(dest, src, w*h);
	impr_dilaten(dest, dest, w, h, conn, times);
	impr_negate(dest, dest, w*h);
}

/**
 * impr_close
 * @dest: destination image
 * @src: source image
 * @w: image width
 * @h: image height
 *
 * "Close" image.
 * @dest == @src is ok.
 */
void impr_close(b8 *dest, b8 *src, int w, int h)
{
	impr_dilate(dest, src, w, h);
	impr_erode(dest, dest, w, h);
}

/**
 * impr_open
 * @dest: destination image
 * @src: source image
 * @w: image width
 * @h: image height
 *
 * "Open" image.
 * @dest == @src is ok.
 */
void impr_open(b8 *dest, b8 *src, int w, int h)
{
	impr_erode(dest, src, w, h);
	impr_dilate(dest, dest, w, h);
}

/**
 * impr_and
 * @dest: destination image
 * @src1: 1st source image
 * @src2: 2nd source image
 * @size: image size
 *
 * Binary AND on each pixel pair of images.
 * @dest == @src1 or @src2 is ok.
 */
void impr_and(b8 *dest, b8 *src1, b8 *src2, int size)
{
	int i;
	for (i=0; i<size; i++)
		dest[i] = src1[i] & src2[i];
}

/**
 * impr_or
 * @dest: destination image
 * @src1: 1st source image
 * @src2: 2nd source image
 * @size: image size
 *
 * Binary OR on each pixel pair of images.
 * @dest == @src1 or @src2 is ok.
 */
void impr_or(b8 *dest, b8 *src1, b8 *src2, int size)
{
	int i;
	for (i=0; i<size; i++)
		dest[i] = src1[i] | src2[i];
}
