#include "impr.h"

/**
 * impr_downscale
 * @dest: destination, smaller image
 * @src: original image
 * @w: image width
 * @h: image height
 * @factor: shrink image by this integer factor
 *
 * Shrinks image by integer @factor. @w and @h have to be multiples
 * of @factor.
 */
void impr_downscale(u8 *dest, u8 *src, int w, int h, int factor)
{
	int i, j;
	int id, jd;
	int wd = w/factor;
	int hd = h/factor;
	w = wd*factor;
	h = hd*factor;

	for (jd=0; jd<hd; jd++)
		for (id=0; id<wd; id++) {
			int pixel_value = 0;
			int pos = (jd*w + id) * factor;

			for (j=0; j<factor; j++) {
				for (i=0; i<factor; i++)
					pixel_value += src[pos++];
				pos += w-factor;
			}
			
			dest[jd*wd + id] = pixel_value/(factor*factor);
		}
}

/**
 * impr_downscale2
 * @dest: destination, smaller image
 * @src: original image
 * @w: image width
 * @h: image height
 * @factorx: shrink image in x dimension by this integer factor
 * @factory: shrink image in y dimension by this integer factor
 *
 * Shrinks image by integers @factorx and @factory.
 * @w and @h have to be multiples of @factorx and @factory respectively.
 */
void impr_downscale2(u8 *dest, u8 *src, int w, int h, int factorx, int factory)
{
	int i, j;
	int id, jd;
	int wd = w/factorx;
	int hd = h/factory;
	w = wd*factorx;
	h = hd*factory;

	for (jd=0; jd<hd; jd++)
		for (id=0; id<wd; id++) {
			int pixel_value = 0;
			int pos = jd*w * factory + id * factorx;

			for (j=0; j<factory; j++) {
				for (i=0; i<factorx; i++)
					pixel_value += src[pos++];
				pos += w-factorx;
			}
			
			dest[jd*wd + id] = pixel_value/(factorx*factory);
		}
}

/* A B *
 * C D */
static inline void _impr_interpolate_abcd(u8 *dest, int pos, int w,
		int factorx, int factory,
		int a, int b, int c, int d)
{
	int sqrf = factorx*factory;

	int ac = factory*a;
	int bd = factory*b;

	int i, j;

	for (j=0; j<factory; j++) {
		int abcd = factorx*ac/*-ac/2 + bd/2*/;

		for (i=0; i<factorx; i++) {
			dest[pos] = abcd / sqrf;

			abcd += -ac + bd;
			pos++;
		}
		pos += w-factorx;

		ac += -a + c;
		bd += -b + d;
	}
}

/**
 * impr_upscale
 * @dest: destination, larger
 * @src: original image
 * @wd: source image width
 * @hd: source image height
 * @factor: widen image by this integer factor
 *
 * Increases image size with linear interpolation.
 * @dw and @dh have to be multiples of @factor.
 * Note: with 2**x factors, we could use shifting, which could be faster.
 */
void impr_upscale(u8 *dest, u8 *src, int wd, int hd, int factor)
{
	int id, jd;
	int w = wd*factor;
	int bord_r, bord_b; /* right and bottom "border" offsets */
	int f2 = factor/2;

	/* center area */
	for (jd=0; jd<hd-1; jd++)
		for (id=0; id<wd-1; id++) {
			int posd = jd*wd + id;
			/* A B *
			 * C D */
			int a = src[posd];
			int b = src[posd+1];
			int c = src[posd+wd];
			int d = src[posd+wd+1];

			_impr_interpolate_abcd(dest, (jd*factor+f2)*w + id*factor+f2,
					w, factor, factor, a, b, c, d);
		}

	bord_r = (wd-1)*factor+f2;
	bord_b = ((hd-1)*factor+f2)*w;
	/* left and right borders */
	for (jd=0; jd<hd-1; jd++) {
		int posd = jd*wd;
		int pos = (jd*factor+f2)*w;
		int a, b, c, d;

		a = b = src[posd];
		c = d = src[posd+wd];
		_impr_interpolate_abcd(dest, pos, w, f2, factor, a, b, c, d);

		posd += wd-1;
		a = b = src[posd];
		c = d = src[posd+wd];
		_impr_interpolate_abcd(dest, pos + bord_r, w, f2, factor, a, b, c, d);
	}

	/* top and bottom borders */
	for (id=0; id<wd-1; id++) {
		int posd = id;
		int pos = id*factor+f2;
		int a, b, c, d;

		a = c = src[posd];
		b = d = src[posd+1];
		_impr_interpolate_abcd(dest, pos, w, factor, f2, a, b, c, d);

		posd += (hd-1)*wd;
		a = c = src[posd];
		b = d = src[posd+1];
		_impr_interpolate_abcd(dest, bord_b + pos, w, factor, f2, a, b, c, d);
	}

	/* and the corners */
	{
		int a;

		a = src[0];
		_impr_interpolate_abcd(dest, 0, w, f2, f2, a, a, a, a);

		a = src[wd-1];
		_impr_interpolate_abcd(dest, bord_r, w, f2, f2, a, a, a, a);

		a = src[(hd-1)*wd];
		_impr_interpolate_abcd(dest, bord_b, w, f2, f2, a, a, a, a);

		a = src[(hd-1)*wd + wd-1];
		_impr_interpolate_abcd(dest, bord_b + bord_r, w, f2, f2, a, a, a, a);
	}
}

/**
 * impr_upscale2
 * @dest: destination, larger
 * @src: original image
 * @wd: source image width
 * @hd: source image height
 * @factorx: widen image in x dimension by this integer factor
 * @factory: widen image in y dimension by this integer factor
 *
 * Increases image size with linear interpolation.
 * @dw and @dh have to be multiples of @factorx and @factory respectively.
 */
void impr_upscale2(u8 *dest, u8 *src, int wd, int hd, int factorx, int factory)
{
	int id, jd;
	int w = wd*factorx;
	int bord_r, bord_b; /* right and bottom "border" offsets */
	int fx2 = factorx/2;
	int fy2 = factory/2;

	/* center area */
	for (jd=0; jd<hd-1; jd++)
		for (id=0; id<wd-1; id++) {
			int posd = jd*wd + id;
			/* A B *
			 * C D */
			int a = src[posd];
			int b = src[posd+1];
			int c = src[posd+wd];
			int d = src[posd+wd+1];

			_impr_interpolate_abcd(dest, (jd*factory+fy2)*w + id*factorx+fx2,
					w, factorx, factory, a, b, c, d);
		}

	bord_r = (wd-1)*factorx+fx2;
	bord_b = ((hd-1)*factory+fy2)*w;
	/* left and right borders */
	for (jd=0; jd<hd-1; jd++) {
		int posd = jd*wd;
		int pos = (jd*factory+fy2)*w;
		int a, b, c, d;

		a = b = src[posd];
		c = d = src[posd+wd];
		_impr_interpolate_abcd(dest, pos, w, fx2, factory, a, b, c, d);

		posd += wd-1;
		a = b = src[posd];
		c = d = src[posd+wd];
		_impr_interpolate_abcd(dest, pos + bord_r, w, fx2, factory, a, b, c, d);
	}

	/* top and bottom borders */
	for (id=0; id<wd-1; id++) {
		int posd = id;
		int pos = id*factorx+fx2;
		int a, b, c, d;

		a = c = src[posd];
		b = d = src[posd+1];
		_impr_interpolate_abcd(dest, pos, w, factorx, fy2, a, b, c, d);

		posd += (hd-1)*wd;
		a = c = src[posd];
		b = d = src[posd+1];
		_impr_interpolate_abcd(dest, bord_b + pos, w, factorx, fy2, a, b, c, d);
	}

	/* and the corners */
	{
		int a;

		a = src[0];
		_impr_interpolate_abcd(dest, 0, w, fx2, fy2, a, a, a, a);

		a = src[wd-1];
		_impr_interpolate_abcd(dest, bord_r, w, fx2, fy2, a, a, a, a);

		a = src[(hd-1)*wd];
		_impr_interpolate_abcd(dest, bord_b, w, fx2, fy2, a, a, a, a);

		a = src[(hd-1)*wd + wd-1];
		_impr_interpolate_abcd(dest, bord_b + bord_r, w, fx2, fy2, a, a, a, a);
	}
}
