#include <string.h>
#include "segmentize.h"
#include "helpers.h"

/*
 * consider this case, which is usually problematic:
 * |--------------|
 * |  |  |  |  |  |
 * |  |  |  |  |
 * |  |  |  |
 * |  |  |
 * |  |
 * |
 *
 * Now, with this code all duplicates are set to minimum
 * so code that fixes them up only traverses once from
 * min to max, and gets all right.
 * You can think of this as adding edges(minv, maxv) to a graph,
 * and in the fixup stage, assigning each connected subgraph
 * a separate (min of vertices) number.
 */
static inline u16 duplicate_add(u16 *duplicates, u16 reg1, u16 reg2)
{
	/* just set them to lowest common region number */
	u16 rmin = min(duplicates[duplicates[reg1]], duplicates[duplicates[reg2]]);

	duplicates[duplicates[reg1]] = rmin;
	duplicates[duplicates[reg2]] = rmin;
	duplicates[reg1] = rmin;
	duplicates[reg2] = rmin;

	return rmin;
}

static inline void duplicate_new(u16 *duplicates, u16 reg)
{
	duplicates[reg] = reg;
}

static inline int _segmentize(u16 *region_mask, b8 *chan,
		int w, int h, int connectedness)
{
	int size = w*h;
	int nregionparts = 1;
	int nregions;
	int i, j;
	static u16 duplicates[1<<16];
	static u16 region_numbers[1<<16];

	// TODO remove debugging
	memset(duplicates, 0xba, sizeof(duplicates));
	memset(region_numbers, 0xbe, sizeof(region_numbers));
	region_numbers[0] = 0;	/* borders */

	/* border guards */
	for (j=0; j<size; j += w)
		chan[j] = region_mask[j] = 0;
	for (i=0; i<w; i++)
		chan[i] = region_mask[i]= 0;

#if 0
/* stairs case, for problematic case of segmentation */
for (j=1; j<50; j++) {
	int pos = j*w;
	for (i=1; i<50; i++) {
		chan[pos+i] = 1;
	}
}
for (j=1; j<50*w; j += w) {
	if (j<40*w)
		chan[j+10] = 0;
	if (j<30*w)
		chan[j+20] = 0;
	if (j<20*w)
		chan[j+30] = 0;
	if (j<10*w)
		chan[j+40] = 0;
}
/* x net, to test 4 vs. 8 connectedness */
for (i=1; i<130; i++)
	for (j=1; j<i; j++)
		chan[(j+i)*w + i-j] = 0;
#endif

	for (j=1; j<h; j++) {
		int pos = j*w + 1;
		for (i=1; i<w; i++) {
			/* regions (region_mask) */
			u16 reg_this;
			u16 reg_left = region_mask[pos-1];
			u16 reg_upper = region_mask[pos-w];
			u16 reg_upperleft, reg_upperright;

			/* bw channel pixels */
			b8 pix_this = chan[pos];
			b8 pix_left = chan[pos-1];
			b8 pix_upper = chan[pos-w];
			b8 pix_upperleft, pix_upperright;

			if (connectedness == 8) {
				reg_upperleft = region_mask[pos-w-1];
				reg_upperright = region_mask[pos-w+1];
				/* ^ this is ok, on i==w-1, it becomes first
				 * pixel of next line, which is border. */
				pix_upperleft = chan[pos-w-1];
				pix_upperright = chan[pos-w+1];
			}

			/* not a part of region (ie. edge) */
			if (!pix_this) {
				region_mask[pos] = 0;
				pos++;
				continue;
			}

			/* check surroundings */
			if (pix_left) {
				reg_this = reg_left;
				/* pixel is a merge point of two regions */
				if (pix_upper)
					reg_this = duplicate_add(duplicates, reg_upper, reg_left);

			} else if (pix_upper) {
				reg_this = reg_upper;

			} else if (pix_upperleft && connectedness == 8) {
				reg_this = reg_upperleft;
				if (pix_upperright)
					reg_this = duplicate_add(duplicates, reg_upperleft, reg_upperright);

			} else if (pix_upperright && connectedness == 8) {
				reg_this = reg_upperright;

			} else { /* no surrounding, this is a new region */
				reg_this = nregionparts++;
				duplicate_new(duplicates, reg_this);
				region_numbers[reg_this] = 0; /* clearing on the fly, as needed */
				assert(nregionparts < (1<<16));
			}

			region_mask[pos] = reg_this;
			pos++;
		}
	}
	duplicates[0] = 0; /* rbefore* without-if fix */

	nregions = 1;
	for (i=1; i<nregionparts; i++) {
		/* fixup merged regions */
		if (duplicates[duplicates[i]])
			duplicates[i] = duplicates[duplicates[i]];

		/* renumber with 1 to nregions */
		if (!region_numbers[duplicates[i]])
			region_numbers[duplicates[i]] = nregions++;
	}

	for (i=0; i<size; i++) {
		u16 reg = region_numbers[duplicates[region_mask[i]]];
		assert (duplicates[region_mask[i]] != 0xbaba);
		assert (reg!=0xbebe);
		region_mask[i] = reg;
	}

	return nregions;
}

/* make two copies of the function */

/**
 * segmentize_8
 * @region_mask: returned regions, [x]=0 for border, numbers enumerate
 * 	regions
 * @chan: segmentize this; [x]=0 are considered borders, 1 region parts
 * @w: image width
 * @h: image height
 *
 * Performs 8-connected region segmentation.
 * Returns number of regions, with 0 to that number in @region_mask.
 */
int segmentize_8(u16 *region_mask, b8 *chan, int w, int h)
{
	return _segmentize(region_mask, chan, w, h, 8);
}

/**
 * segmentize_4
 * @region_mask: returned regions, [x]=0 for border, numbers enumerate
 * 	regions
 * @chan: segmentize this; [x]=0 are considered borders, 1 region parts
 * @w: image width
 * @h: image height
 *
 * Performs 4-connected region segmentation.
 * Returns number of regions, with 0 to that number in @region_mask.
 */
int segmentize_4(u16 *region_mask, b8 *chan, int w, int h)
{
	return _segmentize(region_mask, chan, w, h, 4);
}
